@@ -6,7 +6,12 @@ function (x)
 
 dQuote =
 function(x)
-  paste('"', x, '"', sep = "")
+{
+  if(length(x) == 0)
+     character(0)
+  else
+      paste('"', x, '"', sep = "")
+}
 
 isContainer =
 function(x, asIs, .level)
@@ -17,7 +22,7 @@ setGeneric("toJSON",
   function(x, container = isContainer(x, asIs, .level),
            collapse = "\n", ...,
            .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0,
-            .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA)   {
+            .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity")   {
 
   container; .withNames  # force these values.
   
@@ -34,27 +39,47 @@ setGeneric("toJSON",
 setMethod("toJSON", "NULL",
            function(x, container = isContainer(x, asIs, .level), collapse = "\n", ...,
                      .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null",
-                      .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+                      .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
              if(container) "[ null ] " else "null"
            })
 
+setMethod("toJSON", "array",
+           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ...,
+                       .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0,
+                      .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
+
+             d = dim(x)
+             txt = apply(x, length(d), toJSON, collapse = collapse, ..., .level = .level + 1L, .withNames = .withNames, .na = .na, .escapeEscapes = .escapeEscapes, pretty = pretty, asIs = asIs, .inf = .inf)
+              paste(c("[", paste(txt, collapse = ", "), "]"), collapse = collapse)
+           })
+
+
+setMethod("toJSON", "function",
+           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ...,
+                       .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0,
+                      .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
+             toJSON(paste(deparse(x), collapse = collapse), container, collapse, ..., .level = .level, .withNames = .withNames, .na = .na, .escapeEscapes = .escapeEscapes, pretty = pretty, asIs = asIs, .inf = .inf)
+           })
+
+
 
 setMethod("toJSON", "ANY",
            function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ...,
                        .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0,
-                      .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+                      .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
 
              if(isS4(x)) {
                paste("{", paste(dQuote(slotNames(x)), sapply(slotNames(x),
                                                               function(id)
                                                                  toJSON(slot(x, id), ..., .level = .level + 1L,
-                                                                        .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs)),
+                                                                        .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs,
+                                                                         .inf = .inf)),
                                  sep = ": ", collapse = ","),
                      "}", collapse = collapse)
              } else {
 #cat(class(x), "\n")
                if(is.language(x)) {
-                  return(toJSON(as.list(x), asIs = asIs))
+                  return(toJSON(as.list(x), asIs = asIs, .inf = .inf, .na = .na))
                   stop("No method for converting ", class(x), " to JSON")
                }
 
@@ -71,8 +96,11 @@ setMethod("toJSON", "integer",
            function(x, container =  isContainer(x, asIs, .level),
                       collapse = "\n  ", ..., .level = 1L,
                       .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null",
-                       .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
-
+                       .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity")
+          {
+            if(any(is.infinite(x)))
+              warning("non-fininte values in integer vector may not be approriately represented in JSON")
+            
              if(any(nas <- is.na(x)))
                  x[nas] = .na
 
@@ -90,7 +118,7 @@ setOldClass("hexmode")
 setMethod("toJSON", "hexmode",
            function(x, container =  isContainer(x, asIs, .level), collapse = "\n   ", ...,
                      .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0,
-                       .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+                       .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
 
              
              tmp = paste("0x", format(x), sep = "")
@@ -110,7 +138,7 @@ setMethod("toJSON", "hexmode",
 setMethod("toJSON", "factor",
            function(x, container =  isContainer(x, asIs, .level),
                      collapse = "\n", ..., .level = 1L,
-                     .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", pretty = FALSE, asIs = NA) {
+                     .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", pretty = FALSE, asIs = NA, .inf = " Infinity") {
 
              toJSON(as.character(x), container, collapse, ..., .level = .level, .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs)
            })
@@ -118,7 +146,7 @@ setMethod("toJSON", "factor",
 setMethod("toJSON", "logical",
            function(x, container =  isContainer(x, asIs, .level),
                      collapse = "\n", ..., .level = 1L,
-                     .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", pretty = FALSE, asIs = NA) {
+                     .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", pretty = FALSE, asIs = NA, .inf = " Infinity") {
              tmp = ifelse(x, "true", "false")
              if(any(nas <- is.na(tmp)))
                  tmp[nas] = .na             
@@ -135,16 +163,24 @@ setMethod("toJSON", "logical",
 setMethod("toJSON", "numeric",
            function(x, container =  isContainer(x, asIs, .level), collapse = "\n", digits = 5, ...,
                       .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0,
-                        .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+                        .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA,
+                         .inf = " Infinity") {
+
+            if(any(is.infinite(x)))
+              warning("non-fininte values in numeric vector may not be approriately represented in JSON")             
 
              tmp = formatC(x, digits = digits)
+              
              if(any(nas <- is.na(x)))
-                 tmp[nas] = .na                          
+                 tmp[nas] = .na
+             if(any(inf <- is.infinite(x))) 
+                 tmp[inf] = sprintf(" %s%s", ifelse(x[inf] < 0, "-", ""), .inf)
              
              if(container) {
                 if(.withNames)
-                   paste(sprintf("{%s", collapse), paste(dQuote(names(x)), tmp, sep = ": ", collapse = sprintf(",%s", collapse)),
-                                   sprintf("%s}", collapse))
+                   paste(sprintf("{%s", collapse),
+                         paste(dQuote(names(x)), tmp, sep = ": ", collapse = sprintf(",%s", collapse)),
+                         sprintf("%s}", collapse))
                 else
                    paste("[", paste(tmp, collapse = ", "), "]")
              } else
@@ -153,7 +189,7 @@ setMethod("toJSON", "numeric",
 
 
 setMethod("toJSON", "character",
-           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", digits = 5, ..., .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", digits = 5, ..., .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
 # Don't do this: !             tmp = gsub("\\\n", "\\\\n", x)
 
 #             if(length(x) == 0)    return("[ ]")
@@ -190,7 +226,7 @@ setMethod("toJSON", "character",
 # Symbols.
 # names can't be NA
 setMethod("toJSON", "name",
-           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ..., .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ..., .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
              as.character(x)
            })
 
@@ -203,16 +239,17 @@ setMethod("toJSON", "name",
 
 setOldClass("AsIs")
 setMethod("toJSON", "AsIs",
-           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ..., .level=1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, asIs = NA) {
-              toJSON(structure(x, class = class(x)[-1]), container = TRUE, collapse = collapse, ..., .level = .level + 1L, .withNames = .withNames, .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs)
+           function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ..., .level=1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, asIs = NA, .inf = " Infinity") {
+              toJSON(structure(x, class = class(x)[-1]), container = TRUE, collapse = collapse, ..., .level = .level + 1L, .withNames = .withNames, .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs, .inf = .inf)
            })
 
 
 
 setMethod("toJSON", "matrix",
            function(x, container =  isContainer(x, asIs, .level), collapse = "\n", ...,
-                    .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
-             tmp = paste(apply(x, 1, toJSON, .na = .na, ..., .escapeEscapes = .escapeEscapes), collapse = sprintf(",%s", collapse))
+                    .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity")
+           {
+             tmp = paste(apply(x, 1, toJSON, .na = .na, ..., .escapeEscapes = .escapeEscapes, .inf = .inf), collapse = sprintf(",%s", collapse))
              if(!container)
                return(tmp)
 
@@ -223,7 +260,7 @@ setMethod("toJSON", "matrix",
            })
 
 setMethod("toJSON", "list",
-           function(x, container = isContainer(x, asIs, .level), collapse = "\n", ..., .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
+           function(x, container = isContainer(x, asIs, .level), collapse = "\n", ..., .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null", .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
                 # Degenerate case.
 #browser()
              if(length(x) == 0) {
@@ -231,7 +268,7 @@ setMethod("toJSON", "list",
                 return(if(is.null(names(x))) "[]" else "{}")
              }
 
-             els = lapply(x, toJSON, ..., .level = .level + 1L, .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs)
+             els = lapply(x, toJSON, ..., .level = .level + 1L, .na = .na, .escapeEscapes = .escapeEscapes, asIs = asIs, .inf = .inf)
 
              if(all(sapply(els, is.name)))
                names(els) = NULL
@@ -252,6 +289,33 @@ setMethod("toJSON", "list",
            })
 
 
+if(TRUE)
+setMethod("toJSON", "data.frame",
+           function(x, container = isContainer(x, asIs, .level), collapse = "\n", ..., byrow = FALSE, colNames = FALSE, .level = 1L,
+                    .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null",
+                    .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity")
+ {
+  if(byrow) {
+     tmp = lapply(1:nrow(x), function(i) { row = as.list(x[i, ])
+                                           if(colNames)
+                                             row                                             
+                                           else
+                                             unname(row)
+                                         })
+     toJSON(tmp, container, collapse, ..., .level = .level, .withNames = FALSE,
+            .na = .na, .escapeEscapes = .escapeEscapes, pretty = pretty, asIs = asIs, .inf = .inf)
+  } else
+              # Data frame columns should always be treated as containers. From Joe Cheng
+      toJSON(lapply(as.list(x),
+                    function(col) 
+                       I(col)
+                   ),
+              container = container, collapse = collapse, ...,
+              .level = .level, .withNames = .withNames, .na = .na,
+              .escapeEscapes = .escapeEscapes, pretty = pretty, asIs = asIs, .inf = .inf)
+           })
+
+
 jsonPretty =
 function(txt)
 {
@@ -266,7 +330,7 @@ function(txt)
 setMethod("toJSON", "environment",
            function(x, container = isContainer(x, asIs, .level), collapse = "\n  ", ...,
                      .level = 1L, .withNames = length(x) > 0 && length(names(x)) > 0, .na = "null",
-                      .escapeEscapes = TRUE, pretty = FALSE, asIs = NA) {
-      toJSON(as.list(x), container, collapse, .level = .level, .withNames = .withNames, .escapeEscapes = .escapeEscapes, asIs = asIs)
+                      .escapeEscapes = TRUE, pretty = FALSE, asIs = NA, .inf = " Infinity") {
+      toJSON(as.list(x), container, collapse, .level = .level, .withNames = .withNames, .escapeEscapes = .escapeEscapes, asIs = asIs, .inf = .inf)
            })
 
